package dareka;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Arrays;
import java.util.Map;
import java.util.Properties;

import dareka.common.CloseUtil;
import dareka.common.Config;
import dareka.common.DiskFreeSpace;
import dareka.common.Logger;
import dareka.processor.impl.Cache;

public class Main {
	// version string
	public static String VER_STRING
							= "NicoCache_nl SP1.07 HBsoft-patch1 (based on NicoCache v0.32)";
    public static String getVersion() {
        return VER_STRING;
    }
	
	private static Server server;

	public static void stop() {
		if (server != null) {
			server.stop();
		}
	}
	
    public static void main(String[] args) {
        try {
            mainBody();
        } catch (Exception e) {
            Logger.debug(e);
        }
    }
    
    @SuppressWarnings("deprecation")
    private static void mainBody() throws IOException {
    	// [nl] iniɂlpBłmɂinit@CȂ
        File configFile = new File("config.ini");
        if (configFile.exists() == false)
        	configFile = new File("config.properties");
        
        // ݒt@CȂăftHgݒt@CȂ烊l[Ďg
        if (configFile.exists() == false)
        {
        	File defFile = new File("config.properties.default");
        	if (defFile.exists())
        	{
        		defFile.renameTo(configFile);
        	}
        }

        // ǂݍ
        Config config = configure(configFile);
        
        Logger.info(VER_STRING);
        Logger.info("    Running with Java %s on %s",
        				System.getProperty("java.version"),
        				System.getProperty("os.name"));

        Logger.info("port=" + config.getListenPort());
        if (config.getProxyHost().equals("")) {
        	Logger.info("direct mode (no secondary proxy)");
        } else {
        	Logger.info("proxy host=" + config.getProxyHost());
        	Logger.info("proxy port=" + config.getProxyPort());
        }

        // [nl] IP
    	config.allowFrom = Config.getString("allowFrom", "local");
		if (config.allowFrom.equals("local"))
			Logger.info(" => Only localhost Allowed");
		else if (config.allowFrom.equals("all"))
			Logger.info(" => Anyone can access NicoCache");
		else if (config.allowFrom.startsWith("lan"))
		{
			if (config.allowFrom.equals("lan"))
				config.allowFrom = "lanC";
			Logger.info(" => Only LAN Address can access NicoCache (mode: %s)", config.allowFrom);
		}
		else
			Logger.info(" => Invalid allowFrom setting. Assumes 'local' mode.");
		
		// [nl] ڑ搧
		if (Boolean.getBoolean("niconicoMode"))
			Logger.info(" => Only nico|smilevideo.jp domains are processed.");
		
		// [nl] x
		String limitStr;
		if ((limitStr = System.getProperty("speedLimit")) != null)
		{
			int limit = Integer.parseInt(limitStr);
			if (limit < 8 || 1000 < limit)
				System.clearProperty("speedLimit");
			else
				Logger.info(" => Transfer Speed Limit: " + limitStr + "Mbps");
		}
		
		// LbVɃ^Cgt邩ǂ
        Logger.info("title=" + String.valueOf(config.isTitle()));
                
        // [nl] XNvgߍݐݒ
        if (Integer.parseInt(Config.getString("scriptOn", "0")) != 0)
        {
            config.scriptTarget = Config.getString("scriptTarget", "no_uri_to_replace");
            config.scriptText = Config.getString("scriptText", "");
        }
        else
        	config.scriptTarget = "no_uri_to_replace";
        if (config.scriptTarget.equals("no_uri_to_replace") == false)
        	Logger.info("Script replace On");
        
        // [nl] flvplayer_wrapper
        config.flvWrapper = Config.getString("flvWrapper", "false");
        if (config.flvWrapper.equals("false") == false)
    	{
    		if (config.flvWrapper.equals("true"))
    			Logger.info("flvplayer_wrapper replace On (RC)");
			else if (config.flvWrapper.startsWith("rc2"))
				if (config.flvWrapper.startsWith("rc2_")) {
					String h = config.flvWrapper.substring(4);
					Logger.info("flvplayer_wrapper replace On (RC2 height=" + h + ")");
				} else
					Logger.info("flvplayer_wrapper replace On (RC2 height=510)");
			else
        		config.flvWrapper = "false";
    	}
        
        config.localFlv = Config.getBoolean("localFlv", true);
    	if (config.localFlv)
    		Logger.info("LocalFlv Server On");
    	
        if (Boolean.getBoolean("resumeDownload")) {
            Logger.info("Resume suspended download On");
        }

        if (Boolean.getBoolean("touchCache")) {
            Logger.info("Touch Cache File On");
        }
		
		String ngwordtxt = System.getProperty("NGWORDtxt", "");
		if (ngwordtxt.equals("") == false)
			config.ngword = new File(ngwordtxt);
		if (config.ngword != null)
		{
			if (config.ngword.canWrite())
				Logger.info("Add Proxomitron NGWord On");
			else {
				Logger.info("Invalid Proxomitron NGWord.txt Path or Setting");
				config.ngword = null;
			}
		}
		
        if (Boolean.getBoolean("dareka.debug")) {
            Logger.info("debug mode");
        }
        
		// [nl] BݒBgĂH
		config.printHeader = Config.getBoolean("printHeader", false);
        
        // TODO _E[h֌W
        if (Boolean.getBoolean("continueDownload"))
        	Logger.info("continueDownload is NOT supported.");
        
        // [nl] ȈՐU蕪@\
        config.storeFilter = Config.getBoolean("storeFilter", false);
        if (config.storeFilter)
        	Logger.info("Storing Folder Filter On");

        // [nl] LbVtH_̎w
		config.cacheFolder = Config.getString("cacheFolder", config.cacheFolder);
		if (config.cacheFolder.equals("")) {
			config.cacheFolder = "cache";
			Cache.init(config);
		} else {
			Logger.info("Cache Folder: " + config.cacheFolder);
			Cache.init(config, config.cacheFolder);
		}
        Cache.cleanup();
        Logger.info("total cache size=%,dbytes", Cache.size());
        
        long freeSize = DiskFreeSpace.get(config.cacheFolder);
        if (freeSize != Long.MAX_VALUE) {
        	int neededSize = Config.getInteger("needFreeSpace", 100);
        	Logger.info("cache folder free space = %,dbytes (at least %,dMB)", freeSize, neededSize);
        } else {
        	Logger.info("Can't read disk free size. os.name = '%s'", System.getProperty("os.name"));
        }
        
        if (Config.getBoolean("cacheAllocateFirst", false)) {
        	Logger.info("Allocate cache space before download: On");
        }
        
        // [nl] TlCLbV
		config.cacheThumbnail = Config.getBoolean("cacheThumbnail", config.cacheThumbnail);
        if (config.cacheThumbnail) {
        	// ꎞIɖ (#034);
        	Logger.info("cacheThumbnail is temporarily disabled.");
        	config.cacheThumbnail = false;
        	//ThumbProcessor.init();
        }
        Logger.info("----------");

		server = new Server(config);
		try {
			server.start();
		} catch (RuntimeException e) {
			Logger.error(e);
		}
    }
    
    @SuppressWarnings("deprecation")
    private static Config configure(File configFile)
            throws FileNotFoundException, IOException {
        Properties p = new Properties();
        setDefaults(p);

        Config config = new Config();
        if (configFile.exists()) {
            loadFrom(configFile, p);

            String v;

            v = p.getProperty(Config.LISTEN_PORT);
            if (v != null) {
                config.setListenPort(Integer.parseInt(v));
            }

            v = p.getProperty(Config.PROXY_HOST);
            if (v != null) {
                config.setProxyHost(v);
            }

            v = p.getProperty(Config.PROXY_PORT);
            if (v != null) {
                config.setProxyPort(Integer.parseInt(v));
            }

            v = p.getProperty(Config.TITLE);
            if (v != null) {
                config.setTitle(Boolean.parseBoolean(v));
            }
        } else {
            FileOutputStream out = new FileOutputStream(configFile);
            try { // ensure closing out
                p.store(out, "NicoCache config file");
            } finally {
                CloseUtil.close(out);
            }
        }

        for (Map.Entry<Object, Object> entry : p.entrySet()) {
            String key = (String) entry.getKey();
            String value = (String) entry.getValue();
            System.setProperty(key, value.trim());
        }

        return config;
    }

    private static void loadFrom(File configFile, Properties p)
            throws FileNotFoundException, IOException {
        FileInputStream in = new FileInputStream(configFile);
        try { // ensure closing in
            p.load(in);
        } finally {
            CloseUtil.close(in);
        }
    }

    private static void setDefaults(Properties p) {
        try {
            setDefaultsFromFiles(p);
            if (p.size() > 0) {
                return;
            }
        } catch (IOException e) {
            Logger.debug(e);
        }

        p.setProperty("listenPort", "8080");
        p.setProperty("proxyHost", "");
        p.setProperty("proxyPort", "8081");
        p.setProperty("title", "true");
        p.setProperty("touchCache", "true");
    }

    private static void setDefaultsFromFiles(Properties p) throws IOException {
        File defaultsDir = new File("defaults");

        File[] files = defaultsDir.listFiles(new FilenameFilter() {
            public boolean accept(File dir, String name) {
                if (name.endsWith(".properties")) {
                    return true;
                } else {
                    return false;
                }
            }
        });

        if (files == null) {
            throw new IOException("failed to read: " + defaultsDir);
        }

        Arrays.sort(files);
        for (File f : files) {
            loadFrom(f, p);
        }
    }
}