package dareka.processor.impl;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Collections;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;

import dareka.Main;
import dareka.common.Config;
import dareka.common.Logger;

class CacheStats
{
	public long size;
	public long timestamp;
	public CacheStats(long size, long timestamp)
	{
		this.size = size;
		this.timestamp = timestamp;
	}
}

public class Cache extends CacheOriginal {
	
	// [nl] fBNgXg
	private static Vector<String> dirList = new Vector<String>();
	private static ConcurrentHashMap<String, CacheStats> statCache
						= new ConcurrentHashMap<String, CacheStats>();
	private static Config config = null;
	
    public static void init(Config conf) {
    	config = conf;
    	CacheOriginal.init();
    	dirList.clear();
    	
        searchCachesOnADirectory(cacheDir, 1);
    }
    
    // [nl] LbVtH_w肵
    public static void init(Config conf, String cachePath) {
    	cacheDir = new File(cachePath);
    	init(conf);
    }
    
    //[nl] LbVtH_擾
    public static String getCacheDir() {
    	return cacheDir.getAbsolutePath();
    }
    
	// v0.06
    protected static void searchCachesOnADirectory(File dir, int depth) {
    	// [nl] fBNgꗗ
        dirList.add(getDirName(dir));
        
        File[] cacheFiles = dir.listFiles();
        for (File file : cacheFiles) {
            if (file.isDirectory()) {
                searchCachesOnADirectory(file, depth + 1);
            } else if (file.isFile()) {
                String id = getIdFromFilename(file.getName());
                if (id.length() > 0) {
                	if (depth == 1 && id.equals(NLTMP)) {
                		id = getIdFromFilename(file.getName().substring(6));
                		id2Tmp.put(id, file);
                		
                        Logger.debug("partial cache found: " + id + " => "
                                + file.getPath());
                	} else {
                		id2File.put(id, file);

                		Logger.debug("cache found: " + id + " => " + file.getPath());
                	}
                }
            }
        }
    }
    
    static String getDirName(File dir)
    {
    	String dirname = dir.getPath().substring(cacheDir.getPath().length());
        dirname = dirname.replace(File.separatorChar, '/').replaceFirst("^/", "");
        return dirname;
    }

	// [nl] w肵ID̃LbVʏ킩GRm~[𔻒inullj
	public static String getType(String id)
	{
		if (id2File.get(id) != null)
			return "";
		else if (id2File.get(id + "low") != null)
			return "low";
		return null;
	}

	// [nl] flvlist𐶐ĕԂ
	public static String getFlvList()
	{
		String str = "<html>\n" +
					 "<head>\n" +
					 "\t<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">" +
					 "\t<meta http-equiv=\"Pragma\" content=\"no-cache\">\n" +
					 "\t<meta http-equiv=\"Cache-Control\" content=\"no-cache\">\n" +
					 "\t<meta http-equiv=\"Expires\" content=\"Thu, 01 Dec 1994 16:00:00 GMT\">\n" +
					 "\t<script type=\"text/javascript\" src=\"/local/list.js\"></script>\n" +
					 "\t<link rel=\"stylesheet\" type=\"text/css\" href=\"/local/list.css\" charset=\"utf-8\">\n" +
					 "\t<title>LocalFLV</title>\n" +
					 "</head>\n<body>\n"; 
		str += "<script type=\"text/javascript\">\n";
		str += getFlvListAjax();
		str += "makeCacheList();\n";
		str += "</script>\n</body>\n</html>";
		return str;
	}
	
	public static String getFlvListAjax()
	{
		// ǂݒƂɂ܂BƂ΂
		init(config);
		String str = new String("ncversion = \"" + Main.VER_STRING + "\";\n");
		
		str += "dirList = " + getDirListAsJson(false) + ";\n";
		str += "tempList = " + getTempListAsJson(false) + ";\n";
		str += "cacheList = " + getCacheListAsJson(false) + ";\n";

		// [JFLVflvplayer̃wb_pɔÔ
		if (id2File.size() > 0) {
				StringBuilder forLocal = new StringBuilder("forLocalFlv = \"");
				for (String k : id2File.keySet().toArray(new String[0])) {
					forLocal.append(k).append(".flv ");
				}
				forLocal.append("\";\n");
				str += forLocal;
		}
		return str;
	}
	
	public static String getFlvListAsJson()
	{
		init(config);
		String str = new String("{\n\"dirList\":") +
			getDirListAsJson(false) +
			",\n\"tempList\":" +
			getTempListAsJson(false) +
			",\n\"cacheList\":" +
			getCacheListAsJson(false) +
			"\n}";

		return str;
	}
	
	//[nl] + JSON fix
	//getFlvlistAjax ɂ܂
	public static String getCacheListAsJson(boolean initCache)
	{
		if (initCache)
			init(config);
		StringBuffer strBuf = new StringBuffer("{\n");
				
		if (id2File.size() > 0)
		{
				String[] lst = MovieIdSorter.sort(id2File, false);
				
				for (String k : lst)
				{
					File cFile = id2File.get(k);
					String fn = cFile.getName();
					String path = getDirName(cFile.getParentFile());
					long lastmod, size;
					// ς݃LbV̌ʂ̓LbVĂ݂
					CacheStats entry;
					if ((entry = statCache.get(k)) != null) {
						size = entry.size;
						lastmod = entry.timestamp;
					} else {
						size = cFile.length();
						lastmod = cFile.lastModified() / 1000;
						entry = new CacheStats(size, lastmod);
						statCache.put(k, entry);
					}
					String title = "";
					if (fn.length() > k.length()+5)
						title = fn.substring(k.length()+1, fn.length()-4);
					
					strBuf.append(
						"\"" + k + "\":[\"" + title + "\", \"" + path + "\", " +
							size + ", " + lastmod +"],\n");
				}
			int sblen = strBuf.length();
			strBuf.replace(sblen - 2, sblen, "\n}");
		} else {
			strBuf.append("}");
		}

		return strBuf.toString();
	}

	public static String getCacheListAsJson()
	{
		return getCacheListAsJson(true);
	}

	//[nl] + JSON fix
	public static String getTempListAsJson(boolean initCache)
	{
		if (initCache)
			init(config);
		StringBuffer strBuf = new StringBuffer("{\n");
				
		if (id2Tmp.size() > 0)
		{
			String[] lstTmp = id2Tmp.keySet().toArray(new String[0]);
			for (String k : lstTmp)
			{
				File kFile = id2Tmp.get(k);
				boolean nowDL = getDLFlag(k);
				String fn = kFile.getName().substring(6);
				long lastmod = kFile.lastModified() / 1000;
				long nowSize = kFile.length();
				long realSize = 0;
				if (nowDL)
					realSize = id2DL.get(k);
				String title = (fn.length() > (k.length()+5) ? fn.substring(k.length()+1, fn.length()-4) : "");
				
				strBuf.append("\"" + k + "\":[\"" + title + "\", " + String.valueOf(nowSize) + ", " +
						String.valueOf(realSize) + ", " + String.valueOf(nowDL) + ", " + lastmod +"],\n"); 
			}
			int sblen = strBuf.length();
			strBuf.replace(sblen - 2, sblen, "\n}");
		} else {
			strBuf.append("}");
		}

		return strBuf.toString();
	}

	public static String getTempListAsJson()
	{
		return getTempListAsJson(true);
	}

	//[nl] + JSON fix
	public static String getDirListAsJson(boolean initCache)
	{
		if (initCache)
			init(config);
		
		StringBuffer str = new StringBuffer("[");
		for (int i=0; i<dirList.size(); i++)
		{
			String path = dirList.get(i);
			if (i != 0)
				str.append(", ");
			str.append("\"").append(path).append("\"");
		}
		str.append("]");

		return new String(str);
	}

	public static String getDirListAsJson() {
		return getDirListAsJson(true);
	}
	
	// [nl] wrapperp̃Xg
	public static String getLocalFlvList()
	{
		StringBuilder list = new StringBuilder();
		for (String k : id2File.keySet())
			list.append(k).append(".flv\r\n");
		return list.toString();
	}
	
	// [nl] tH_Ԉړ
	public boolean moveTo(String dir)
	{
		if (dirList.contains(dir))
		{
			File toDir = new File(cacheDir.getPath()+"/" + dir, cacheFile.getName());
			if (cacheFile.renameTo(toDir))
			{
				id2File.put(cacheId, toDir);
				return true;
			}
		}
		return false;
	}
	
	// [nl] LbV菜
	public static boolean remove(String id)
	{
		File cache = id2File.get(id);
        if (cache != null)
		{
			if (cache.exists())
			{
				cache.delete();
				id2File.remove(id);
				return true;
			}
		}
        return false;
	}

	// [nl] ꎞt@C菜
	public static boolean removeTmp(String id)
	{
		File tmp = id2Tmp.get(id);
        if (tmp != null && getDLFlag(id) == false)
		{
			if (tmp.exists())
				tmp.delete();
			id2Tmp.remove(id);
			return true;
		}
        return false;
	}

	// RXgN^
	public Cache(String cacheId) {
		super(cacheId, "v"); //TODO:b
	}
	
	public Cache(String cacheId, String extension) {
		super(cacheId, extension);
	}

	// [nl] ^Cg擾iꍇIDj
    public String getTitle()
    {
    	String filename = this.getCacheFileName();
        Matcher m = CACHE_FILE_PATTERN.matcher(filename);
        if (m.find()) {
        	if (m.group(2) != null)
        		return m.group(2);
        	else
        		return m.group(1);
        } else {
            return "";
        }
    }
	
    // [nl] t@C̃^CgύX
    public boolean setTitle(String newTitle)
    {
    	File newName = getDescribedCacheFile(getId(), newTitle);
    	newName = new File(cacheFile.getParentFile(), newName.getName());
    	if (cacheFile.renameTo(newName))
    	{
    		id2File.put(getId(), newName);
    		cacheFile = newName;
    		return true;
    	}
    	return false;
    }
    
    /**
     * [nl] _E[h̐
     */
    public static int getDLFlagCount() {
    	return id2DL.size();
    }
    
    private class DirInfo implements Comparable<DirInfo>
    {
    	public String path;
    	public String name;
    	public int depth;
    	public DirInfo(String path, String name, int depth) {
    		this.path = path;
    		this.name = name;
    		this.depth = depth;
    	}
    	
    	// OE[
    	public int compareTo(DirInfo e) {
    		if (this.name.length() == e.name.length())
    			return this.depth == e.depth ? 0 : (this.depth < e.depth ? 1 : -1);
    		return this.name.length() < e.name.length() ? 1 : -1;
    	}
    }
    
    // [nl]tH_tB^ƂāAXgAT
    protected String suggestStoreFolder()
    {
    	Vector<DirInfo> dirs = new Vector<DirInfo>();
    	for (String dir: dirList)
    	{
    		String[] parts = dir.split("/");
    		int n = parts.length;
    		dirs.add(new DirInfo(dir, tidyTitle(parts[n-1].toLowerCase()), n));
    	}
    	// Ƀ\[gāA "/" Ă
    	Collections.sort(dirs);
    	dirs.remove(dirs.size() - 1);
    	
    	String test = cacheFile.getName().toLowerCase();
    	for (DirInfo entry: dirs) {
    		String[] patterns = entry.name.split(" +");
    		boolean match = true;
    		for (String p: patterns) {
    			if (p.charAt(0) == '-') {
    				// NOT
    				if (test.contains(p.substring(1)))
    					match = false;
    			} else {
    				// AND
    				if (!test.contains(p))
    					match = false;
    			}
    		}
    		if (match)
    			return entry.path;
    	}
    		
    	return "";
    }
    
    // LbVK؂ȈʒuEOŊi[B
    // [nl] GRm~[łLΏ悤ɁBU蕪֌Wǉ
    //		GRm~[ł΂̏ꏊɕۑ悤 (#027, #028)
    public void store() throws IOException {
        File cacheTmpFile = getCacheTmpFile();

        File parentDir = cacheDir;
        if (!cacheId.endsWith("low")) {
            String lowId = cacheId + "low";
        	File lowFile = id2File.get(lowId);
        	if (lowFile != null) { 
        		parentDir = lowFile.getParentFile();
        		Cache.remove(lowId);
        	}
        }
        
        if (parentDir == cacheDir && config.storeFilter) {
        	String storeFolder = suggestStoreFolder();
        	if (!storeFolder.equals("")) {
        		Logger.info("storing folder: " + storeFolder);
        		parentDir = new File(cacheDir, storeFolder);
        	}
        }
        cacheFile = new File(parentDir, cacheFile.getName());
        
        if (cacheTmpFile.renameTo(cacheFile)) {
            id2File.put(cacheId, cacheFile);
        } else {
            cacheTmpFile.delete();
        }
        id2Tmp.remove(cacheId);
    }
    
    // [nl] TODO:ЂƂ܂ꎞt@C폜
    public static void cleanup()
    {
    	// do nothing
    }
}
