package dareka.processor.impl;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.Socket;
import java.net.URLDecoder;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Config;
import dareka.common.Logger;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.URLResource;
// [nl] cachetH_֌W̏
public class CacheDirProcessor implements Processor {
    private static final String[] SUPPORTED_METHODS = new String[] { "GET" };
    private static final Pattern CACHE_DIR_PATTERN = Pattern
    	.compile("^http://www\\.nicovideo\\.jp/cache/?(|.+)$");
    
    public String[] getSupportedMethods() {
		return SUPPORTED_METHODS;
	}

	public Pattern getSupportedURLAsPattern() {
		return CACHE_DIR_PATTERN;
	}

	public String getSupportedURLAsString() {
		return null;
	}

	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		String path = requestHeader.getPath();
		
		if (path.length() < 7)
			return null;
		path = path.substring(7);
		
		// [nl] [JFLViLbV𑗂j
		Pattern LOCAL_FLV_PATTERN = Pattern.compile("^([a-z]{2}[0-9]+)(?:low)?\\.(flv|swf|mp4)$");
		Matcher mLocalPath = LOCAL_FLV_PATTERN.matcher(path);
		if (mLocalPath.lookingAt()) {
			String cacheId = mLocalPath.group(1);
			Cache cache = new Cache(cacheId);
			if (cache.exists() == false) {
				cache = new Cache(cacheId + "low");
			}
			
			if (cache.exists()) {
				Logger.info("Local Flv: " + cache.getId() + " " + cache.getTitle());
				if (Boolean.getBoolean("touchCache")) {
					cache.touch();
				}
				Resource r = Resource.get(Resource.Type.URL, cache.getURLString());
				String extension = cache.getPostfix();
				String contentType = "video/flv";
				if (extension.equals(Cache.SWF)) {
					contentType = "application/x-shockwave-flash";
				} else if (extension.equals(Cache.MP4)) {
					contentType = "video/mp4";
				}
				r.addResponseHeader(HttpHeader.CONTENT_TYPE, contentType);
				r.addResponseHeader("Content-Disposition", "attachment");
				r.addTransferListener(new LimitFlvSpeedListener());
				return r;
			}
			else {
				return StringResource.getNotFound();
			}
		}
		// mp3ۑ
		else if (path.matches("[a-z]{2}[0-9]+(low)?\\.mp3")) {
			String cacheId = path.substring(0, path.length()-4);
			Cache cache = new Cache(cacheId);
			if (!cache.exists()) {
				cache = new Cache(cacheId += "low");
			}
			
			// FLVȊO͒ołȂ̂ŃG[Ԃ
			if (!cache.getPostfix().equals(Cache.FLV)) {
				Logger.info("MP3 Converter: Only FLV movies supported.");
				return StringResource.getNotFound();
			}
			
			if (cache.exists()) {
				Logger.info("MP3 Converter: " + cache.getId() + " " + cache.getTitle());
				return new Flv2mp3(cache);
			}
			return StringResource.getNotFound();
		}
		// m4aۑ
		else if (path.matches("[a-z]{2}[0-9]+\\.m4a")) {
			String cacheId = path.substring(0, path.length()-4);
			Cache cache = new Cache(cacheId);
			if (!cache.exists()) {
				return StringResource.getNotFound();
			}
			
			// MP4ȊO͒ołȂ̂ŃG[Ԃ
			if (!cache.getPostfix().equals(Cache.MP4)) {
				Logger.info("M4A Converter: Only MP4 movies supported.");
				return StringResource.getNotFound();
			}
			
			// MP4BoxƒołȂ̂ŃG[Ԃ
			if (!(new File("MP4Box.exe").exists())) {
				Logger.info("Needs `MP4Box.exe' to extract audio.");
				return StringResource.getNotFound();
			}
			
			File m4atmp = new File(Cache.getCacheDir(), "_m4atmp.mp4");
			int i = 0;
			while(m4atmp.exists()) {
				m4atmp = new File(Cache.getCacheDir(), "_m4atmp" + (++i) + ".mp4");
			}
			m4atmp = new File(Cache.getCacheDir(), m4atmp.getName().replace(".mp4", ""));
			
			//StringBuilder command = new StringBuilder("MP4Box ");
			//command.append("\"").append(cache.getCacheFile().getAbsolutePath()).append("\"");
			//command.append(" -single 2");
			//command.append(" -out \"").append(m4atmp.getAbsolutePath()).append("\"");
			
			// [GUI] Runtime#exec() ̌ĂяoύXĂ݂
			String[] commands = new String[6];
			commands[0] = "MP4Box";
			commands[1] = cache.getCacheFile().getAbsolutePath();
			commands[2] = "-single";
			commands[3] = "2";
			commands[4] = "-out";
			commands[5] = m4atmp.getAbsolutePath();
			
			Logger.debug("MP4Box  in:" + commands[1]);
			Logger.debug("MP4Box out:" + commands[5]);
			
			// [GUI] Runtime#exec() ̌ĂяoύXĂ݂
			//Process proc = Runtime.getRuntime().exec(command.toString());
			Process proc = Runtime.getRuntime().exec(commands);
			proc.getOutputStream().close();
			
			BufferedReader inputReader = new BufferedReader(new InputStreamReader(
					proc.getInputStream()));
			while(inputReader.readLine() != null) {
				
			}

			try {
				proc.waitFor();
				proc.getErrorStream().close();
				inputReader.close();
				proc.destroy();
			} catch (InterruptedException e) {
			}
			
			// [GUI] ÂMP4Box͏o̓t@CႤ߁Ãt@CmF
			File m4aout;
			m4aout = new File(Cache.getCacheDir(), m4atmp.getName() + ".mp4");
			if (m4aout.exists()) {
				Logger.info("M4A Converter: " + cache.getId() + " " + cache.getTitle());
				Resource r = new TemporaryFileResource(m4aout);
				r.setResponseHeader("Content-Type", "audio/m4a");
				r.setResponseHeader("Content-Disposition",
									"attachment; filename=\"" + path + "\"");
				return r;
			}
			m4aout = new File(Cache.getCacheDir(), m4atmp.getName() + "_track2.mp4");
			if (m4aout.exists()) {
				Logger.info("M4A Converter: " + cache.getId() + " " + cache.getTitle());
				Resource r = new TemporaryFileResource(m4aout);
				r.setResponseHeader("Content-Type", "audio/m4a");
				r.setResponseHeader("Content-Disposition",
									"attachment; filename=\"" + path + "\"");
				return r;
			}
			return StringResource.getNotFound();
		}
		// [nl] crossdomain.xmlNotFoundihCȂ̂Ŗ{KvȂj
		else if (path.equals("crossdomain.xml")) {
			return StringResource.getNotFound();
		}
		// [nl] LbVǗy[W
		else if (path.equals("")) {
			// [nl] swf̃NGXgȂȈՔłɂ
			if (requestHeader.getMessageHeader("x-flash-version") != null) {
				String flvList = "# my cache\r\n" + Cache.getLocalFlvList() + "\n";
				StringResource r = new StringResource(flvList);
				r.addNoCacheResponseHeaders();
				return r;
			}			
			long start = System.currentTimeMillis();
			String str = Cache.getFlvList();
			long elapsed = System.currentTimeMillis() - start;
			Logger.info("Local Flv List (" + elapsed + "msec)");
			StringResource r = new StringResource(str);
			r.addNoCacheResponseHeaders();
			return r;
		}
		// [nl] [hȂflvidꗗ (FxłLocal Flvp)
		else if (path.equals("flvlist")) {
			String flvList = "# my cache\r\n" + Cache.getLocalFlvList() + "\n";
			StringResource r = new StringResource(flvList);
			r.addNoCacheResponseHeaders();
			return r;
		}
		// [nl] LbVǗy[W(ajax)
		else if (path.equals("ajax")) {
			long start = System.currentTimeMillis();
			String str = Cache.getFlvListAjax();
			long elapsed = System.currentTimeMillis() - start;
			Logger.info("Local Flv List ajax (" + elapsed + "msec)");
			StringResource r = new StringResource(str);
			r.addNoCacheResponseHeaders();
			return r;
		}
		//[nl] + JSON fix 
		else if (path.endsWith(".json")) {
			String name = null;
			String str = null;
			long start = System.currentTimeMillis();
			if (path.equals("cachelist.json")) {
				name = "Local Cache List as JSON";
				str = Cache.getCacheListAsJson();
			} else if (path.equals("templist.json")) {
				name = "Local Temp Cache List as JSON";
				str = Cache.getTempListAsJson();
			} else if (path.equals("dirlist.json")) {
				name = "Local Cache Dir List as JSON";
				str = Cache.getDirListAsJson();
			} else if (path.equals("flvlist.json")) {
				name = "Local Flv List as JSON";
				str = Cache.getFlvListAsJson();
			}
			long elapsed = System.currentTimeMillis() - start;

			if (str != null) {
				Logger.info("%s (%d msec)", name, elapsed);
				StringResource r = new StringResource(str);
				r.addNoCacheResponseHeaders();
				return r;
			}
			return StringResource.getNotFound();
		}
		//[nl] + JSON fix ܂

		// fR[hAajaxǂ̔
		path = URLDecoder.decode(path, "UTF-8");
		boolean ajax = false;
		if (path.startsWith("ajax_")) {
			ajax = true;
			path = path.substring(5);
		}

		String msg = "NG";
		// [nl] Xgy[W
		// [nl] [JFLVXg폜
		if (path.matches("rm(tmp)?\\?([a-z]{2}[0-9]+(low)?)"))
		{
			if (path.startsWith("rmtmp"))
			{
				String smid = path.substring(6);
				Logger.info("Remove Temporary: " + smid);
				if (Cache.removeTmp(smid) == false)
					Logger.info("...failed.");
				else
					msg = "OK";
			}
			else
			{
				String smid = path.substring(3);
				Logger.info("Remove Cache: " + smid);
				if (Cache.remove(smid))
					msg = "OK";
			}
		}
		// [nl] [JFLVXg^Cg擾
		else if (path.matches("title\\?[a-z]{2}[0-9]+(low)?"))
		{
			String cacheId = path.substring(6);
			NicoCachingTitleRetriever titleRetriever
				= new NicoCachingTitleRetriever(cacheId, requestHeader);
			try {
				String title = titleRetriever.call();
				if (title != null)
				{
					Cache cache = new Cache(cacheId);
					if (cache.exists())
					{
						cache.setTitle(title);
						msg = "OK " + Cache.getSanitizedDescription(title);
					}
				}
			} catch (Exception e) {
				msg = "NG";
			}
		}
		// [nl] [JFLVXg^Cgݒ
		else if (path.matches("title\\?([a-z]{2}[0-9]+(low)?)-.*"))
		{
			String[] t = path.split("-", 2);
			String cacheId = t[0].substring(6);
			String title = URLDecoder.decode(t[1], "UTF-8");
			if (title != null)
			{
				Cache cache = new Cache(cacheId);
				if (cache.exists())
				{
					cache.setTitle(title);
					msg = "OK " + Cache.getSanitizedDescription(title);
				}
			}
		}
		// [nl] [JFLVXgtH_ړ
		else if (path.matches("move\\?([a-z]{2}[0-9]+(low)?)-.*"))
		{
			String[] t = path.split("-", 2);
			String cacheId = t[0].substring(5);
			Cache cache = new Cache(cacheId);
			if (cache.moveTo(t[1]))
				msg = "OK";
		}
		else
		{
			Logger.info("no method:" + path);
			return null;
		}
		
		// nhLꍇ͍ŌɉԂ
		if (ajax) {
			StringResource r = new StringResource(msg);
			r.addNoCacheResponseHeaders();
			return r;
		} else {
			String redir = null;
			String nlFrom = requestHeader.getParameter("nl-region");
			if (nlFrom != null
					&& (nlFrom.equals("cache") || nlFrom.equals("local")))
				redir = requestHeader.getParameter("nl-from");
			else
				redir = "http://www.nicovideo.jp/";
			StringResource r = StringResource.getRedirect(redir);
			r.addNoCacheResponseHeaders();
			return r;
		}
	}
	
	// MIt@C폜Resource
	class TemporaryFileResource extends URLResource {
		File fileToDelete = null;
		public TemporaryFileResource(File deleteFile) throws IOException {
			super(deleteFile.toURI().toString());
			fileToDelete = deleteFile;
		}
		
		@Override
	    public boolean endEnsuredTransferTo(Socket receiver,
	            HttpRequestHeader requestHeader, Config config) throws IOException {
	        boolean result = super.endEnsuredTransferTo(receiver, requestHeader, config);
			if (fileToDelete.exists()) {
				fileToDelete.delete();
			}
	        return result;
	    }
	}
}