package dareka.processor.impl;

import java.io.IOException;
import java.util.regex.Pattern;

import dareka.common.Logger;
import dareka.extensions.RequestFilter;
import dareka.processor.HttpRequestHeader;

public class DefaultRequestFilter implements RequestFilter {
	private static final String allowDomains =
		"^(?:" +
		"(.+\\.)?nicovideo\\.jp" +
		"|(.+\\.)?smilevideo\\.jp" +
		"|local\\.ptron" +
		"|(|g-)ecx\\.images-amazon\\.com" +
		"|(.+\\.)?nicovideo\\.photozou\\.jp" +
		")$";
	private static final Pattern allowDomainPattern
		= Pattern.compile(allowDomains);
	
	public boolean isDenyDomain(String host) {
		// [nl] niconicoMode
		if (Boolean.getBoolean("niconicoMode")) {
			boolean allow = allowDomainPattern.matcher(host).matches();
			if (allow == false) {
				Logger.info("Forbidden host access: " + host);
				return true;
			}
		}
		return false;
	}

	public int onRequest(HttpRequestHeader requestHeader) throws IOException {
		// niconicoMode
		if (isDenyDomain(requestHeader.getHost()))
			return RequestFilter.DROP;
		
		// Referer𒼂
		String referer = requestHeader.getMessageHeader("Referer");
		if (referer != null) {
			if (referer.startsWith("http://www.nicovideo.jp/cache")) {
				requestHeader.setMessageHeader("Referer", "http://www.nicovideo.jp/my");
				requestHeader.setParameter("nl-from", referer);
				requestHeader.setParameter("nl-region", "cache");
			} else if (referer.startsWith("http://www.nicovideo.jp/local")) {
				requestHeader.setMessageHeader("Referer", "http://www.nicovideo.jp/my");
				requestHeader.setParameter("nl-from", referer);
				requestHeader.setParameter("nl-region", "local");
			}
		}
		return RequestFilter.OK;
	}
}
