package dareka.processor.impl;

import java.io.OutputStream;

import dareka.common.Config;
import dareka.processor.HttpResponseHeader;
import dareka.processor.TransferListener;

// xdListener
public class LimitFlvSpeedListener implements TransferListener
{
	private long transferedBytes = 0;
	private long startedTime = 0;
	private long limitMbps = 1000;
	public LimitFlvSpeedListener() {
		limitMbps = Config.getInteger("speedLimit", 1000);
	}
	
	public LimitFlvSpeedListener(long mbps) {
		limitMbps = mbps;
	}
	
	// TransferListener interface
	public void onResponseHeader(HttpResponseHeader responseHeader) { }

	public void onTransferBegin(OutputStream receiverOut) {
		startedTime = System.currentTimeMillis();
	}
	
	public void onTransferring(byte[] buf, int length) {
		try {
			transferedBytes += length;
			long elapsedTime = System.currentTimeMillis() - startedTime;
			while (elapsedTime > 0 && transferedBytes / elapsedTime > limitMbps * 1024 / 8) {
				Thread.sleep(3);
				elapsedTime = System.currentTimeMillis() - startedTime;
			}
		} catch (InterruptedException e) {
		}
	}
	
	public void onTransferEnd(boolean completed) {
		//Logger.info("" + (System.currentTimeMillis() - startedTime));
		//Logger.info("kbps: " + 8 * transferedBytes / (System.currentTimeMillis() - startedTime));
	}
}