package dareka.processor.impl;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.Socket;
import java.util.concurrent.Executor;
import java.util.concurrent.FutureTask;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.CloseUtil;
import dareka.common.Logger;
import dareka.processor.HttpHeader;
import dareka.processor.HttpRequestHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;

public class NicoCachingProcessor implements Processor {
    private static final String[] SUPPORTED_METHODS = new String[] { "GET" };
    /**
     * SMILEVIDEO̓URL̐K\
     */
    // http://s-clb5.smilevideo.jp/smile?v=102982.92382
    private static final Pattern SM_FLV_PATTERN =
            Pattern.compile("^http://[^/]+(?:smilevideo|nicovideo)\\.jp/smile\\?([msv])=([^.]+)\\.\\d+(low)?$");

    private Executor executor;

    public NicoCachingProcessor(Executor executor) {
        this.executor = executor;
    }

    public String[] getSupportedMethods() {
        return SUPPORTED_METHODS;
    }

    public Pattern getSupportedURLAsPattern() {
        return SM_FLV_PATTERN;
    }

    public String getSupportedURLAsString() {
        return null;
    }

    public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
            throws IOException {
        Matcher m = SM_FLV_PATTERN.matcher(requestHeader.getURI());
        if (!m.find()) {
            // it must not happen...
            return Resource.get(Resource.Type.URL, requestHeader.getURI());
        }        

        String fm = m.group(1);
        String id = m.group(2);
        String smid;
        String suffix = m.group(3);
        Cache cache;
        
        // [nl] ^CgLbVɂΈꎞt@Cɂ^Cgt (#025)
        //		smȊOŎn܂IDɂΉ
        String title = null;
        if (NicoCachingTitleRetriever.containsId(id)) {
        	smid = NicoCachingTitleRetriever.getSMID(id);
        	title = NicoCachingTitleRetriever.getTitleBySmid(smid);
        } else {
        	// NGXg琄
        	smid = "sm" + id;
        	String cookie = requestHeader.getMessageHeader("Cookie");
        	if (cookie != null) {
        		Pattern HISTORY_PATTERN = Pattern.compile("nicohistory=([a-z]{2}(\\d++))");
        		Matcher mHistory = HISTORY_PATTERN.matcher(cookie);
        		if (mHistory.find() && mHistory.group(2).equals(id)) {
        			smid = mHistory.group(1);
        		}
        	}
        }

    	cache = new Cache(smid, fm);
        if (!cache.exists() && suffix != null) {
            cache = new Cache(smid + suffix, fm);
        }

        if (cache.exists()) {
            Logger.info("using cache: " + cache.getCacheFileName());
            if (Boolean.getBoolean("touchCache")) {
                cache.touch();
            }
            Resource r = Resource.get(Resource.Type.URL, cache.getURLString());
            r.setResponseHeader(HttpHeader.CONTENT_TYPE, "video/flv");
            r.addTransferListener(new LimitFlvSpeedListener());

            return r;
        }
        
        // [nl] _E[hꍇ̂݃^CgZbg (#027)
        FutureTask<String> retrieveTitlteTask = null;
        if (Boolean.getBoolean("title")) {
        	if (title != null) {
            	cache.setDescribe(title);
            } else {
            	retrieveTitlteTask =
                    new FutureTask<String>(new NicoCachingTitleRetriever(smid, requestHeader));
            	executor.execute(retrieveTitlteTask);
            }
        }

        Logger.info("no cache found: " + cache.getCacheFileName());

        // [nl] uEŨW[ƁANCł̃W[Ή
        requestHeader.removeMessageHeader("Range");
        requestHeader.removeMessageHeader("If-Range");
        long tmpSize = cache.tmpLength();
        InputStream cacheInput = null;
        // [nl] dDLɂȂꍇ́A2߂̃RlNVł͎Ă镪̃LbV𗘗pB
        // ܂A2߂̃RlNV͕ۑȂBAIɃW[
        boolean dupconnect = Cache.getDLFlag(cache.getId());
        
        // [nl] ꎞLbVL΃W[s
        if (tmpSize != 0
                && (dupconnect || Boolean.getBoolean("resumeDownload"))) {
            tmpSize = cache.tmpLength();
            requestHeader.setMessageHeader("Range", "bytes="
                    + tmpSize + "-");
            cacheInput = new BufferedInputStream(cache.getTmpInputStream());
        }
        
        Resource r;
        try { // ensure cacheInput.close() in error cases.
            // [nl] DLXgɓ
            Cache.setDLFlag(cache.getId(), Integer.MAX_VALUE);

            r = Resource.get(Resource.Type.URL, requestHeader.getURI());

            r.addTransferListener(new NicoCachingListener(cache,
                    retrieveTitlteTask, cacheInput, dupconnect));
        } catch (RuntimeException e) {
            Logger.error(e);
            CloseUtil.close(cacheInput);
            Cache.setDLFlag(cache.getId(), -1);
            throw e;
        }

        return r;
    }
}
