package dareka.processor.impl;

import java.io.ByteArrayOutputStream;
import java.util.concurrent.Callable;
import java.util.concurrent.ConcurrentHashMap;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Logger;
import dareka.common.Pair;
import dareka.processor.HttpRequestHeader;
import dareka.processor.URLResource;

public class NicoCachingTitleRetriever implements Callable<String> {
	// [nl] WatchRewriterŎ擾^CgLbVĂ (#025, #030, rc2.04)
	private static ConcurrentHashMap<String, Pair<String, String>> id2title
						= new ConcurrentHashMap<String, Pair<String, String>>();
	
	public static void putTitleCache(String smid, String title) {
		id2title.put(smid.substring(2),
					 new Pair<String, String>(smid, title));
	}
	
	public static boolean containsId(String id) {
		return id2title.containsKey(id);
	}
	
	public static String getSMID(String id) {
		return id2title.get(id).first;
	}
	
	public static String getTitleBySmid(String smid) {
		return getTitleById(smid.substring(2));
	}
	
	public static String getTitleById(String id) {
		return id2title.get(id).second;
	}
	
    /**
     * ^Cg̐K\
     */
    private static final Pattern HTML_H1_PATTERN =
            Pattern.compile("<h1[^>]*?>(?:<a[^>]*?>|)(.*?)(?:</a>|)</h1>");
    private static final Pattern API_TITLE_PATTERN
			= Pattern.compile("<title>(.+)</title>");

    private String smid;

    public NicoCachingTitleRetriever(String smid, HttpRequestHeader reqHdr) {
        this.smid = smid;
    }

    public String call() throws Exception {
        Logger.debugWithThread("title retrieving start");
        
        // [nl] ^CgLbVɂ΂Ԃ (#025)
        if (containsId(smid.substring(2))) {
        	return getTitleById(smid.substring(2));
        }

        // y[WnlɒʂĂȂꍇAPIŎ擾(rc2.04)
        String url = "http://www.nicovideo.jp/api/getthumbinfo/" + smid;
        URLResource r = new URLResource(url);

        ByteArrayOutputStream bout = new ByteArrayOutputStream();
        r.transferTo(null, bout, null, null);

        String responseString = new String(bout.toByteArray(), "UTF-8");
        String title = getTitleFromXml(responseString);
        
        // 폜APIŎ擾łȂꍇwatch擾(rc2.04)
        if (title == null) {
        	url = "http://www.nicovideo.jp/watch/" + smid;
        	r = new URLResource(url);
        	bout.reset();
            r.transferTo(null, bout, null, null);
            responseString = new String(bout.toByteArray(), "UTF-8");
            title = getTitleFromResponse(responseString);
      	}
        Logger.debugWithThread("title retrieving end (" + title + ")");

        return title;
    }

    /**
     * HTML^Cg擾B
     * 
     * @param htmlSource
     * @return ^CgBȂꍇnullB
     */
    public static String getTitleFromResponse(String htmlSource) {
        Matcher mTitle = HTML_H1_PATTERN.matcher(htmlSource);
        if (mTitle.find()) {
            return mTitle.group(1);
        } else {
            return null;
        }
    }

    /**
     * APIxml^Cg擾B
     * 
     * @param xmlSource
     * @return ^CgBȂꍇnullB
     */
    public static String getTitleFromXml(String xmlSource) {
        Matcher mTitle = API_TITLE_PATTERN.matcher(xmlSource);
        if (mTitle.find()) {
            return mTitle.group(1);
        } else {
            return null;
        }
    }
}
