package dareka.processor.impl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.net.Socket;
import java.util.Map;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Config;
import dareka.common.Logger;
import dareka.common.Pair;
import dareka.extensions.Rewriter;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.URLResource;

public class RewriterProcessor implements Processor {
	private static final String[] SUPPORTED_METHODS = new String[] { "GET",
			"POST"};
	
	protected Config config;
	protected Vector<Rewriter> rewriterList = new Vector<Rewriter>();
	
	public RewriterProcessor(Config config)
	{
		this.config = config;

		// ftHgRewriter
		WatchRewriter wr = new WatchRewriter(config);
		addRewriter(wr);
		ScriptRewriter scr = new ScriptRewriter(config);
		addRewriter(scr);
		ToolbarRewriter tr = new ToolbarRewriter(config);
		addRewriter(tr);
		
		// eXgp̊gRewriter
		//extensions.nlThumbInfoRewriter sr = new extensions.nlThumbInfoRewriter();
		//addRewriter(sr);
	}
	
	public void addRewriter(Rewriter rewriter)
	{
		rewriterList.add(rewriter);
	}
	
	// Processor interface
	public String[] getSupportedMethods() {
		return SUPPORTED_METHODS;
	}
	
	private static final Pattern SUPPORTED_PATTERN = Pattern
			.compile("^http://(?:www|res)\\.nicovideo\\.jp/.*$");

	public Pattern getSupportedURLAsPattern() {
		return SUPPORTED_PATTERN;
	}

	public String getSupportedURLAsString() {
		return null;
	}

	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		String uri = requestHeader.getURI();
		Vector<Pair<Rewriter,Matcher>> matchList = new Vector<Pair<Rewriter,Matcher>>(0);
		
		// URLŒuΏۂL邩`FbN
		for (Rewriter rw: rewriterList)
		{
			Matcher m = rw.getRewriterSupportedURLAsPattern().matcher(uri);
			if (m.lookingAt()) {
				matchList.add(new Pair<Rewriter, Matcher>(rw, m));
			}
		}
		
		// uΏۂΈU_E[hRewriterŏ
		if (matchList.size() > 0)
		{
			//Logger.info(requestHeader.getURI() + " " + requestHeader.getContentLength());
			
			// TODO kΉR[h
			requestHeader.removeHopByHopHeaders();
			requestHeader.removeMessageHeader("Accept-Encoding");
			//requestHeader.addMessageHeader("Connection", "close");
			
			// RecM
			URLResource urlReq = new URLResource(requestHeader.getURI());
			ByteArrayOutputStream bout = new ByteArrayOutputStream();
			urlReq.transferTo(browser.getInputStream(), bout, requestHeader, config);
			
			// xMȂ悤
			requestHeader.setContentLength(0);
			
			// wb_̃p[XƖ{̂̎擾
			ByteArrayInputStream bin = new ByteArrayInputStream(bout.toByteArray());
			HttpResponseHeader responseHeader = new HttpResponseHeader(bin);
			byte[] bcontent = new byte[bin.available()];
			bin.read(bcontent, 0, bin.available());
						
			// eLXgn.php͈ꉞΏ
			String contentType = responseHeader.getMessageHeader("Content-Type");
			if (responseHeader.getStatusCode() == 200 && (
					uri.contains(".php") ||
					(contentType != null && contentType.startsWith("text/"))))
			{
				// nlp[^̃Rs[(rc2.03)
				for (Map.Entry<String,String> entry : requestHeader.getParameters().entrySet()) {
					responseHeader.setParameter(entry.getKey(), entry.getValue());
				}

				//long time = -System.nanoTime();
				String content = new String(bcontent, "UTF-8");
				for (Pair<Rewriter, Matcher> entry: matchList) {
					Rewriter rewriter = entry.first;
					content = rewriter.onMatch(entry.second, responseHeader, content);
					//Logger.info(entry.first.getClass().getName());
				}
				//time += System.nanoTime();
				//Logger.info("Processing time: %,d %s", time, requestHeader.getURI());

				StringResource r = new StringResource(content);
				responseHeader.removeHopByHopHeaders();
				r.copyResponseHeaders(responseHeader);
				return r;
			}
			else {
				// wb_ႤƔ肵͉̂ςȂőM
				responseHeader.removeMessageHeader("Vary");
				responseHeader.removeMessageHeader("Accept-Ranges");
				
				return StringResource.getRawResource(responseHeader, bcontent);
				//return new RawResource(responseHeader, bcontent);
			}
		}
		
		// uΏۂłȂ΂̂܂
		return Resource.get(Resource.Type.URL, requestHeader.getURI());
	}
}
