package dareka.processor.impl;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.net.Socket;
import java.nio.ByteBuffer;
import java.util.Map;
import java.util.Set;
import java.util.Vector;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicLong;
import java.util.regex.Pattern;

import dareka.common.Logger;
import dareka.processor.HttpRequestHeader;
import dareka.processor.HttpResponseHeader;
import dareka.processor.Processor;
import dareka.processor.Resource;
import dareka.processor.StringResource;
import dareka.processor.TransferListener;
import dareka.processor.URLResource;

// [nl] TlC摜LbVĂI
public class ThumbProcessor implements Processor, TransferListener {
    private static final String[] SUPPORTED_METHODS = new String[] { "GET" };
    /**
     * SMILEVIDEO̓URL̐K\
     */
    // http://s-clb5.smilevideo.jp/smile?v=102982.92382
    private static final Pattern SM_THUMB_PATTERN = Pattern
            .compile("^http://tn-skr\\.smilevideo\\.jp/smile\\?i=\\d+$");
    
    private static int SIGNITURE = 1203160629;
    private static int VERSION = 2;
    
    private static ConcurrentHashMap<Integer, ByteBuffer> id2thum = 
    						new ConcurrentHashMap<Integer, ByteBuffer>();
    private static Vector<Integer> newList = new Vector<Integer>();
    private static AtomicLong savedRequest = new AtomicLong(); 
    private static AtomicLong savedTransfer = new AtomicLong();
    
    private static File thcache1 = new File("thcache.bin");
    private static File thcache  = new File("thcache.dat");
    private static boolean readfailed = false;
    
    // TlCLbV̓ǂݍ
    public static void init()
    {
    	if (thcache.exists())
    		init_v2();
    	else if (thcache1.exists())
    	{
    		// v1ȂRo[g
        	init_v1();
        	readfailed = true;
        	store(false);
        	Logger.info("converted v2 format.");
    	}
    }
    
    // format version 2
    private static void init_v2()
    {
    	RandomAccessFile raf = null;
    	try {
			raf = new RandomAccessFile(thcache, "r");
			int signiture = raf.readInt();
			int version = raf.readInt();
			if (signiture != SIGNITURE || version != VERSION)
			{
				readfailed = true;
				return ;
			}

			int size = raf.readInt();
			savedRequest.set(raf.readLong());
			savedTransfer.set(raf.readLong());
			for (int i=0; i<size; i++)
			{
				int key = raf.readInt();
				int len = raf.readInt();
				if (!(0 < len && len < 100000))
					throw new IOException();
				ByteBuffer buf = ByteBuffer.allocate(len);
				raf.readFully(buf.array());
				id2thum.put(key, buf);
			}
		} catch (FileNotFoundException e) {
			readfailed = true;
		} catch (IOException e) {
			// sĂǂ߂Ă镪͎g͂
			readfailed = true;
		}
		finally {
			try {
				raf.close();
			} catch (IOException e) { }
		}
		Logger.info("Thumbnail cache@v2: %,d files (saved %,d requests, %,d bytes)",
				id2thum.size(), savedRequest.get(), savedTransfer.get());
    }
    
    // format version 1
    private static void init_v1()
    {
    	ObjectInputStream ois = null;
    	boolean failed = false;
    	try {
    		InputStream fis = new BufferedInputStream(new FileInputStream(thcache1));
			ois = new ObjectInputStream(fis);
			int size = ois.readInt();
			savedRequest.set(ois.readLong());
			savedTransfer.set(ois.readLong());
			for (int i=0; i<size; i++)
			{
				int key = ois.readInt();
				int len = ois.readInt();
				if (!(0 < len && len < 100000))
					break;
				ByteBuffer buf = ByteBuffer.allocate(len);
				ois.readFully(buf.array());
				id2thum.put(key, buf);
			}
		} catch (IOException e) {
			failed = true;
			Logger.info("ĂI");
			// sĂǂ߂Ă镪͎g͂
		}
		finally
		{
			try {
				if (ois != null)
					ois.close();
			}
			catch (IOException e1) { }
			
			if (failed)
			{
				/*if (thcache.exists())
					thcache.delete();*/
			}
		}
		Logger.info("Thumbnail cache: %,d files (saved %,d requests, %,d bytes)",
						id2thum.size(), savedRequest.get(), savedTransfer.get());
    }
    
    // v2
    protected static void store(boolean silent)
    {
    	Set<Map.Entry<Integer, ByteBuffer>> kvSet = id2thum.entrySet();
    	RandomAccessFile raf = null;
    	boolean writefailed = false;
    	try {
    		raf = new RandomAccessFile(thcache, "rw");
    		raf.writeInt(SIGNITURE);
    		raf.writeInt(VERSION);
			raf.writeInt(kvSet.size());
			raf.writeLong(savedRequest.get());
			raf.writeLong(savedTransfer.get());
			if (readfailed) {
				// SƂ
				for (Map.Entry<Integer, ByteBuffer> entry : kvSet)
				{
					raf.writeInt(entry.getKey());
					byte[] data = entry.getValue().array();
					raf.writeInt(data.length);
					raf.write(data, 0, data.length);
				}
				readfailed = false;
			} else {
				// ǋL̂Ƃ
				raf.seek(raf.length());
				for (Integer key: newList)
				{
					raf.writeInt(key);
					byte []data = id2thum.get(key).array();
					raf.writeInt(data.length);
					raf.write(data, 0, data.length);
				}
				if (!silent)
					Logger.info("%d thumbs added", newList.size());
			}
		} catch (FileNotFoundException e) {
			Logger.warning("Unable to store " + thcache.getName());
		} catch (IOException e) {
			writefailed = true;
			Logger.warning("Storing thumb cache is failed!");
		}
		finally
		{
			newList.clear();
			try {
				if (raf != null)
					raf.close();
			} catch (IOException e1) { }
			if (writefailed)
			{
				if (thcache.exists())
					thcache.delete();
			}
		}
    }
    
    protected static void store_v1(boolean silent)
    {
    	Set<Map.Entry<Integer, ByteBuffer>> kvSet = id2thum.entrySet();
    	ObjectOutputStream o = null;
    	boolean failed = false;
    	try {
    		//RandomAccessFile raf = new RandomAccessFile(thcache, "r");
    		OutputStream fos = new BufferedOutputStream(new FileOutputStream(thcache1));
			o = new ObjectOutputStream(fos);
			o.writeInt(kvSet.size());
			o.writeLong(savedRequest.get());
			o.writeLong(savedTransfer.get());
			for (Map.Entry<Integer, ByteBuffer> entry : kvSet)
			{
				o.writeInt(entry.getKey());
				byte[] data = entry.getValue().array();
				o.writeInt(data.length);
				o.write(data, 0, data.length);
			}
		} catch (FileNotFoundException e) {
			Logger.warning("Unable to open \"thcache.bin\"");
		} catch (IOException e) {
			failed = true;
			Logger.warning("Storing \"thcache.bin\" is failed!");
		}
		finally
		{
			try {
				if (o != null)
					o.close();
			} catch (IOException e1) { }
			if (failed)
			{
				if (thcache.exists())
					thcache.delete();
			}
			else
				if (!silent)
					Logger.info("%,d thumbnails stored.", kvSet.size());
		}
    }

    // Processor̎
    Integer id;
	public String[] getSupportedMethods() {
		return SUPPORTED_METHODS;
	}

	public Pattern getSupportedURLAsPattern() {
		return SM_THUMB_PATTERN;
	}

	public String getSupportedURLAsString() {
		return null;
	}

	public Resource onRequest(HttpRequestHeader requestHeader, Socket browser)
			throws IOException {
		id = Integer.parseInt(requestHeader.getURI().substring(36));
		
		if (id2thum.containsKey(id))
		{
			StringResource r = null;
			if (requestHeader.getMessageHeader("If-Modified-Since") != null)
			{
				savedRequest.incrementAndGet();
				r = StringResource.getNotModified();
			}
			else
			{
				byte[] data = id2thum.get(id).array();
				savedTransfer.addAndGet(data.length);
				long sR = savedRequest.incrementAndGet();

				synchronized (ThumbProcessor.class) {
					if (sR % 20 == 0)
						store(true);
				}
				r = new StringResource(data);
			}
			r.setResponseHeader("Content-Type", "image/jpeg");
			return r;
		}
		URLResource r = new URLResource(requestHeader.getURI());
		r.addTransferListener(this);
		return r;
	}

	// TransferListener̎
	private ByteArrayOutputStream out = null;
	public void onResponseHeader(HttpResponseHeader responseHeader) {
		if (responseHeader.getStatusCode() == 200)
			out = new ByteArrayOutputStream();
		//else
			//Logger.info(responseHeader.getReason() + " " + responseHeader.getStatusCode());
	}

	public void onTransferBegin(OutputStream receiverOut) { }

	public void onTransferring(byte[] buf, int length) {
		if (out != null)
			out.write(buf, 0, length);
	}

	public void onTransferEnd(boolean completed) {
		synchronized (ThumbProcessor.class)
		{
			if (completed && out != null)
			{
				id2thum.put(id, ByteBuffer.wrap(out.toByteArray()));
				newList.add(id);
				if (newList.size() >= 30)
					store(false);
				out = null;
			}
		}
	}
}
