package dareka.processor.impl;

import java.io.File;
import java.io.IOException;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import dareka.common.Config;
import dareka.common.Logger;
import dareka.extensions.Rewriter;
import dareka.processor.HttpResponseHeader;

// [nl] watchy[W
public class WatchRewriter implements Rewriter {
    
    Config config;
    public WatchRewriter(Config config) {
    	this.config = config;
    }

	// Rewriter interface
    private static final Pattern VIDEO_ID_PATTERN
    	= Pattern.compile("video_id = '([a-z]{2}\\d+)';");
	private static final Pattern MOVIE_TYPE_PATTERN
		= Pattern.compile("so\\.addVariable\\(\"movie_type\", \"(\\w+)\"\\);");
    private static final Pattern WATCH_PAGE_PATTERN
    	= Pattern.compile("^http://www\\.nicovideo\\.jp/watch/([a-z]{2}[0-9]+|\\d+)(\\?|\\?(?:lo|eco)=1)?$");
	
	// URL̐K\
	public Pattern getRewriterSupportedURLAsPattern() {
		return WATCH_PAGE_PATTERN;
	}
	
	private static final Pattern WATCH_COMMENT_PATTERN
		= Pattern.compile("<p style=\\\"font-size:12px; line-height:1\\.25;\\\">.+?</p>");
	
	public String onMatch(Matcher match, HttpResponseHeader responseHeader, String content)
			throws IOException {
		String smid = match.group(1);
		
		// [nl] }C̏ꍇ͓eIDĂ (#026)
		if (smid.matches("\\d+")) {
			Matcher m = VIDEO_ID_PATTERN.matcher(content);
			if (m.find()) {
				smid = m.group(1);
			}
		}
		
		// [nl] html^Cg擾 NicoCachingTitleRetriever ɓn (rc2.04)
		String title = NicoCachingTitleRetriever.getTitleFromResponse(content);
		if (title != null) {
			NicoCachingTitleRetriever.putTitleCache(smid, title);
		}
		
		// [nl] ҃Rg֗
		Matcher m = WATCH_COMMENT_PATTERN.matcher(content);
		if (m.find()) {
			String com = m.group(0);
			if (!com.contains("<br />")) {
				com = com.replaceAll("[@]{2,}", "<br />");
			}
			
			com = com.replaceAll("(?<!\")https?://(?!www\\.nicovideo\\.jp)[-\\&\\w\\d.:/%_~?=]+", "<a href=\"$0\" target=\"_blank\">$0</a>")
					 .replaceAll("watch/\\d{10}", "<a class=\"video\" href=\"$0\">$0</a>");
			if (!Config.getFlag("disableRewriterFeatures", "popup", false)) {
				com = popupThumbs(com);
			}
			if (!Config.getFlag("disableRewriterFeatures", "color", false)) {
				com = replaceVideoLink(com);
			}
			
			content = content.substring(0, m.start()) + com + content.substring(m.end()); 
		}
		
		// [nl] SMILEVIDEOACRƓt̉s𖳂 (#028, SP1.01)
		content = content.replaceFirst("(?s)(?:<p>)(<img.+?>)</p>(.*?)(<p class=\"TXT12\">)\\s+(<strong>\\d{4}N)", "$2$3$1 $4");
        
		// [nl] flvplayer_wrapper
        String flvWrapper = config.flvWrapper;
        if (flvWrapper.equals("false") == false)
		{
			if ((new File("local", "flvplayer_wrapper.swf")).exists() != true) {
				Logger.info("local/flvplayer_wrapper.swf ̂Œus܂ł");
			} else {
				//[nl] dlύXɎbΏ(rc2.01)
				String swfo_str = "(new SWFObject\\(\")http://(?:res|www)\\.nicovideo\\.jp/(swf/nicoplayer|flvplayer)(\\.swf\\?ts=([0-9]+)\", \"flvplayer\", \"952\", )\"540\"";
				Pattern SWFO_PATTERN = Pattern.compile(swfo_str);
				Matcher ms = SWFO_PATTERN.matcher(content);
				if (ms.find()) {
					String height = "480";
					String uri_pre= "http://res.nicovideo.jp/";
					if (flvWrapper.startsWith("rc2")) {
						uri_pre = "";
						if (flvWrapper.equals("rc2")) {
							height = "510";
						}
						if (flvWrapper.startsWith("rc2_")) {
							height = flvWrapper.substring(4);
						}
					}
					content = ms.replaceFirst("$1"+uri_pre+"flvplayer_wrapper$3\"" + height + "\"");
					content = content.replaceFirst("height:540px;", "height:" + height + "px;");
				}
			}
		}
        
		// [nl] [JFLVꗗւ̃N
		if (config.localFlv) {
			// LbVƂlowǂ擾BΒʏƂ
			String type = Cache.getType(smid);
			if (type == null) {
				type = "";
			}
			
			// y[Wgq擾
			String extension = Cache.FLV;
			Matcher mType = MOVIE_TYPE_PATTERN.matcher(content);
			if (mType.find()) {
				extension = "." + mType.group(1);
			}
			// LbV̊gq擾
			Cache cache = new Cache(smid + type);
			if (cache.exists()) {
				extension = cache.getPostfix();
			}
			String save = "<a href=\"/cache/" + smid + type + extension + "\">ۑ</a>";
			
			// FLV̂mp3_E[h\
			if (extension.equals(Cache.FLV)) {
				save += "E<a href=\"/cache/" + smid + type + ".mp3\"></a>";
			} else if (extension.equals(Cache.MP4)) {
				if (new File("MP4Box.exe").exists()) {
					save += "E<a href=\"/cache/" + smid + type + ".m4a\"></a>";
				}
			}
			
			// g΃Rg_E[hNǉ
			if (Boolean.getBoolean("extension.nlCommentDownloader")) {
				save += "E<a href=\"/cache/commentext?" + smid + "\">R</a>";
			}
			
			// ߍ
			content = content.replaceFirst(
						"history\">ŋߌ</a>",
						"$0 | <a href=\"/cache/\">LbV</a> (" + save + "E<a href=\"/cache/rm?" + smid + "\">폜</a>)");
		}
		return content;
	}
	
	// static method
	static final Pattern VIDEO_LINK_PATTERN = Pattern.compile("href=\"(?:|/|http://www\\.nicovideo\\.jp/)watch/([a-z]{2}\\d+)\"");
	static final Pattern TO_POPUP_PATTERN = Pattern.compile("href=\"(?:|/|http://www\\.nicovideo\\.jp/)(watch|mylist)/(\\d+|[a-z]{2}\\d+)\"");
	
	public static String popupThumbs(String comment) {
		StringBuffer comBuffer = new StringBuffer("<script type=\"text/javascript\" src=\"/local/overlib_mini.js\"></script>\n");
		comBuffer.append("<script type=\"text/javascript\" src=\"/local/popThumb.js\"></script>\n");
		Matcher popMatch = TO_POPUP_PATTERN.matcher(comment);
		while (popMatch.find()) {
			String targetFolder = "thumb";
			if (popMatch.group(1).equals("mylist")) {
				targetFolder = "thumb_mylist";
			}
			popMatch.appendReplacement(comBuffer, "$0 onMouseOver=\"return popThumb('" + targetFolder + "/$2', 0)\" onMouseOut=\"return nd()\"");
		}
		popMatch.appendTail(comBuffer);
		return new String(comBuffer);
	}
	
	public static String replaceVideoLink(String comment) {
		StringBuffer comBuffer = new StringBuffer();
		Matcher vlMatch = VIDEO_LINK_PATTERN.matcher(comment);
		while (vlMatch.find()) {
			if (Cache.getType(vlMatch.group(1)) != null) {
				vlMatch.appendReplacement(comBuffer, "$0 style=\"color:#C00000;\"");
			}
		}
		vlMatch.appendTail(comBuffer);
		return new String(comBuffer);
	}
}
